/*=========================================================================

Test amoeba optimize with kappa metrics

=========================================================================*/
#if defined(_MSC_VER)
#pragma warning ( disable : 4786 )
#endif


#include <string>
#include <vector>
#include "itkImageRegistrationMethod.h"
										 
#include "itkKappaStatisticImageToImageMetric.h"

#include "itkNearestNeighborInterpolateImageFunction.h"

#include "itkAmoebaOptimizer.h"

#include "itkImage.h"
#include "itkChangeInformationImageFilter.h" 

#include "itkCenteredSimilarity2DTransform.h"
#include "itkCenteredTransformInitializer.h"

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

#include "itkResampleImageFilter.h"
#include "itkCastImageFilter.h"
#include "itkSquaredDifferenceImageFilter.h"


#include "itkImageMaskSpatialObject.h"

//
#include "itkCommand.h"

// observer for AmoebaOptimizer
class CommandIterationUpdateI : public itk::Command
{
public:
  typedef  CommandIterationUpdateI  Self;
  typedef  itk::Command             Superclass;
  typedef itk::SmartPointer<Self>   Pointer;
  itkNewMacro( Self );
protected:
  CommandIterationUpdateI() {};
public:
  typedef itk::AmoebaOptimizer                OptimizerType;
  typedef   const OptimizerType *             OptimizerPointer;

  int  GetIterrationNumber() const
    {
    return IterationNumber;
    }
  void Execute(itk::Object *caller, const itk::EventObject & event)
    {
    Execute( (const itk::Object *)caller, event);
    }

  void Execute(const itk::Object * object, const itk::EventObject & event)
    {
    OptimizerPointer optimizer =
                     dynamic_cast< OptimizerPointer >( object );
    if( ! itk::IterationEvent().CheckEvent( &event ) )
      {
      return;
      }
	OptimizerType::ParametersType parameters = 
                    optimizer->GetCachedCurrentPosition();
	int const taille(parameters.size());
	
	//print
	itk::SingleValuedCostFunction::MeasureType Value;

	Value = optimizer->GetCachedValue();
 

	if (count>=5) {
		count=0;
		std::cout << IterationNumber			 << "   ";
		std::cout << Value       << "   ";
		std::cout << parameters << std::endl;
	   }
	count++;
	IterationNumber++;
   }
private:
	int IterationNumber;
	int count;
};




//main
int main( int argc, char *argv[] )
{

  if( argc < 3 || argc > 3)
    {
    std::cerr << "Missing Parameters " << std::endl;
    std::cerr << "Usage: " << argv[0] ;
	std::cerr << " fixedImageMaskFile movingImageMaskFile";
    std::cerr << std::endl;
    return EXIT_FAILURE;
    }
  /***************************************************************************************
   *                                                                                     *
   *                                                                                     *
   *                                DEFINITION DES TYPES                                 *
   *                                                                                     *
   *                                                                                     *
   ***************************************************************************************/
  // images proprieties
  const    unsigned int    Dimension = 2;

  typedef itk::Image< unsigned char, Dimension >   MaskImageType;

  // transform
  typedef itk::CenteredSimilarity2DTransform< double >   TransformType;

  // optimizer
  typedef itk::AmoebaOptimizer                           OptimizerTypeI;

  //metrics
  typedef itk::KappaStatisticImageToImageMetric< 
                                    MaskImageType, 
                                    MaskImageType >     MetricTypeI;

  

  //interpolation
  typedef itk::NearestNeighborInterpolateImageFunction< 
                                    MaskImageType,
                                    double          >    InterpolatorTypeI;


  //registration
  typedef itk::ImageRegistrationMethod< 
                                    MaskImageType, 
                                    MaskImageType >      RegistrationTypeI;


  //create objects
  MetricTypeI::Pointer          metricI         =  MetricTypeI::New();
  OptimizerTypeI::Pointer       optimizerI      =  OptimizerTypeI::New();
  InterpolatorTypeI::Pointer    interpolatorI   =  InterpolatorTypeI::New();
  TransformType::Pointer        transformI      =  TransformType::New();   
  RegistrationTypeI::Pointer    registrationI   =  RegistrationTypeI::New();
												   

  //links
  //first registration 
  registrationI->SetMetric(        metricI        );
  registrationI->SetOptimizer(     optimizerI     );
  registrationI->SetInterpolator(  interpolatorI  );
  registrationI->SetTransform(     transformI     );


  /***************************************************************************************
   *                                                                                     *
   *                                                                                     *
   *                        LECTURE ET PREPARATION DES MASQUES                           *
   *                                                                                     *
   *                                                                                     *
   ***************************************************************************************/


  typedef itk::ImageFileReader< MaskImageType >    MaskReaderType;

//*******************   Fixed mask	*************************************************
  MaskReaderType::Pointer  fixedMaskReader = MaskReaderType::New();

  fixedMaskReader->SetFileName( argv[1] );

  try 
    { 
    fixedMaskReader->Update(); 
    } 
  catch( itk::ExceptionObject & err ) 
    { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
    } 
  
  typedef itk::ChangeInformationImageFilter < MaskImageType > MaskSpacingFilterType;
  MaskSpacingFilterType::Pointer fixedMaskSpacingFilter = MaskSpacingFilterType::New();
	
  fixedMaskSpacingFilter->SetInput( fixedMaskReader->GetOutput() );

  // set spacing to 1
  MaskImageType::SpacingType FixedMaskSpacing;
  FixedMaskSpacing[0] = 1; // spacing along X
  FixedMaskSpacing[1] = 1; // spacing along Y
  fixedMaskSpacingFilter->SetOutputSpacing( FixedMaskSpacing );
  fixedMaskSpacingFilter->ChangeSpacingOn();

  // set pixel origin to zero
  MaskImageType::PointType FixedMaskOrigin;
  FixedMaskOrigin[0] = 1;  //MovingSpacing[0]/2; 
  FixedMaskOrigin[1] = 1;  //MovingSpacing[0]/2; 
  fixedMaskSpacingFilter->SetOutputOrigin(  FixedMaskOrigin	 );
  fixedMaskSpacingFilter->ChangeOriginOn();
  
  fixedMaskSpacingFilter->Update();
  
  typedef itk::ImageMaskSpatialObject< Dimension >   MaskType;
  MaskType::Pointer  fixedSpatialObjectMask = MaskType::New();

  MaskImageType::Pointer   fixedMask =	fixedMaskSpacingFilter->GetOutput();

  fixedSpatialObjectMask->SetImage( fixedMask );



//************************** moving mask  *******************************
  MaskReaderType::Pointer  movingMaskReader = MaskReaderType::New();

  movingMaskReader->SetFileName( argv[2] );

  try 
    { 
    movingMaskReader->Update(); 
    } 
  catch( itk::ExceptionObject & err ) 
    { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
    } 

  typedef itk::ChangeInformationImageFilter < MaskImageType > MaskSpacingFilterType;
  MaskSpacingFilterType::Pointer movingMaskSpacingFilter = MaskSpacingFilterType::New();


  movingMaskSpacingFilter->SetInput( movingMaskReader->GetOutput() );

  // set spacing to 1
  MaskImageType::SpacingType MovingMaskSpacing;
  MovingMaskSpacing[0] = 1; // spacing along X
  MovingMaskSpacing[1] = 1; // spacing along Y
  movingMaskSpacingFilter->SetOutputSpacing( MovingMaskSpacing );
  movingMaskSpacingFilter->ChangeSpacingOn();

  // set pixel origin to zero
  MaskImageType::PointType MovingMaskOrigin;
  MovingMaskOrigin[0] = 1;  //MovingSpacing[0]/2; 
  MovingMaskOrigin[1] = 1;  //MovingSpacing[0]/2; 
  movingMaskSpacingFilter->SetOutputOrigin(  MovingMaskOrigin	 );
  movingMaskSpacingFilter->ChangeOriginOn();
  
  movingMaskSpacingFilter->Update();


  typedef itk::ImageMaskSpatialObject< Dimension >   MaskType;
  MaskType::Pointer  movingSpatialObjectMask = MaskType::New();

  MaskImageType::Pointer	movingMask = movingMaskSpacingFilter->GetOutput();

  movingSpatialObjectMask->SetImage( movingMaskSpacingFilter->GetOutput() );


  
  OptimizerTypeI::ParametersType initializerII(transformI->GetNumberOfParameters());
  int taille( 2+ initializerII.size() );


	
	  /***************************************************************************************
	   *                                                                                     *
	   *                                                                                     *
	   *                       PARAMETRE ET LANCE LA PREMIERE REGISTRATION                   *
	   *                                 (mask)                                              *
	   *                                                                                     *
	   ***************************************************************************************/
	  
	  //********************** liens  registration  *****************************************

	  // ****  registration **************************
	  registrationI->SetFixedImage(    fixedMask   );
	  registrationI->SetMovingImage(   movingMask  );

	  
	  // ****** metric *******************************
	  metricI->SetForegroundValue(1);
	  metricI->ComplementOn();
										 
	  
	  // ****** transform ***********************************
	  // centre
	  typedef itk::CenteredTransformInitializer< 
										TransformType, 
										MaskImageType, 
										MaskImageType >  TransformInitializerType;
	  TransformInitializerType::Pointer initializerI = TransformInitializerType::New();

	  initializerI->SetTransform(   transformI );
	  initializerI->SetFixedImage(  fixedMask );
	  initializerI->SetMovingImage( movingMask );

	  initializerI->MomentsOn();

	  initializerI->InitializeTransform();
	  
	  // angle
	  transformI->SetScale( 1.0 );

	  //chelle
	  transformI->SetAngle( 0.0 );

	  // lien
	  registrationI->SetInitialTransformParameters( transformI->GetParameters() );


	  //********** optimizer **************************************
	  
	  // chelle
	  typedef OptimizerTypeI::ScalesType       OptimizerScalesType;
	  OptimizerScalesType optimizerIScales( transformI->GetNumberOfParameters() );
	  const double translationScale = 1.0 ;

	  optimizerIScales[0] = 20.0;
	  optimizerIScales[1] = 5.0;
	  optimizerIScales[2] =  1000;
	  optimizerIScales[3] =  1000;
	  optimizerIScales[4] =  translationScale;
	  optimizerIScales[5] =  translationScale;

	  optimizerI->SetScales( optimizerIScales );


	  // parametres
	  OptimizerTypeI::ParametersType  simplexDelta( transformI->GetNumberOfParameters() );
	  simplexDelta[0] = 1.0 ;
	  simplexDelta[1] = 1.0  ;
	  simplexDelta[2] = 1.0  ;
	  simplexDelta[3] = 1.0  ;
	  simplexDelta[4] = 5.0  ;
	  simplexDelta[5] = 5.0  ;

	  optimizerI->AutomaticInitialSimplexOff();
	  optimizerI->SetInitialSimplexDelta( simplexDelta );

	  optimizerI->SetParametersConvergenceTolerance( 0.2 ); 
	  optimizerI->SetFunctionConvergenceTolerance(0.5); 


	  //************** Observer**************************
	  CommandIterationUpdateI::Pointer observerI = CommandIterationUpdateI::New();
	  optimizerI->AddObserver( itk::IterationEvent(), observerI );

	  //************************* Start registration **********************
	  registrationI->Initialize();

	  try 
		{
		registrationI->Update();
		std::cout << "Optimizer stop condition = "
				  << registrationI->GetOptimizer()->GetStopConditionDescription()
				  << std::endl;
		} 
	  catch( itk::ExceptionObject & err ) 
		{ 
		std::cerr << "ExceptionObject caught !" << std::endl; 
		std::cerr << err << std::endl; 
		return EXIT_FAILURE;
		}  

	   // get last parameters (initialization for registrationII


  OptimizerTypeI::ParametersType finalParameters = 
                    registrationI->GetLastTransformParameters();

  // adds final parameters to the file
  std::vector<double> TlineI(taille,0);

  TlineI[1] = metricI->GetValue(finalParameters);

  for(int i=2; i<taille;++i){
	  TlineI[i]=	finalParameters[i-2];
  }
  std::cout <<  TlineI[0] << "   " << TlineI[1] << "  [";
  std::cout <<  TlineI[2] << ", ";
  std::cout <<  TlineI[3] << ", ";
  std::cout <<  TlineI[4] << ", " ;
  std::cout <<  TlineI[5] << ", " ;
  std::cout <<  TlineI[6] << ", " ;
  std::cout <<  TlineI[7] << std::endl;
  
  
  
  return EXIT_SUCCESS;
}